//
//  ConversationMainStatusTableViewCell.swift
//  Tusker
//
//  Created by Shadowfacts on 8/28/18.
//  Copyright © 2018 Shadowfacts. All rights reserved.
//

import UIKit
import Pachyderm

class ConversationMainStatusTableViewCell: UITableViewCell, PreferencesAdaptive {

    var delegate: StatusTableViewCellDelegate? {
        didSet {
            contentLabel.navigationDelegate = delegate
        }
    }
    
    @IBOutlet weak var displayNameLabel: UILabel!
    @IBOutlet weak var usernameLabel: UILabel!
    @IBOutlet weak var contentLabel: StatusContentLabel!
    @IBOutlet weak var avatarImageView: UIImageView!
    @IBOutlet weak var timestampLabel: UILabel!
    @IBOutlet weak var attachmentsView: UIView!
    @IBOutlet weak var favoriteButton: UIButton!
    @IBOutlet weak var reblogButton: UIButton!
    
    var statusID: String!
    var accountID: String!
    
    var favorited: Bool = false {
        didSet {
            DispatchQueue.main.async {
                self.favoriteButton.tintColor = self.favorited ? UIColor(displayP3Red: 1, green: 0.8, blue: 0, alpha: 1) : self.tintColor
            }
        }
    }
    var reblogged: Bool = false {
        didSet {
            DispatchQueue.main.async {
                self.reblogButton.tintColor = self.reblogged ? UIColor(displayP3Red: 1, green: 0.8, blue: 0, alpha: 1) : self.tintColor
            }
        }
    }
    
    var avatarURL: URL?
    var updateTimestampWorkItem: DispatchWorkItem?
    
    override func awakeFromNib() {
        displayNameLabel.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(accountPressed)))
        displayNameLabel.isUserInteractionEnabled = true
        usernameLabel.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(accountPressed)))
        usernameLabel.isUserInteractionEnabled = true
        avatarImageView.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(accountPressed)))
        avatarImageView.isUserInteractionEnabled = true
        avatarImageView.layer.masksToBounds = true
        attachmentsView.layer.cornerRadius = 5
        attachmentsView.layer.masksToBounds = true
    }
    
    func updateUIForPreferences() {
        guard let account = MastodonCache.account(for: accountID) else { fatalError("Missing cached account \(accountID!)") }
        
        avatarImageView.layer.cornerRadius = Preferences.shared.avatarStyle.cornerRadius(for: avatarImageView)
        displayNameLabel.text = account.realDisplayName
    }
    
    func updateUI(for statusID: String) {
        self.statusID = statusID
        
        guard let status = MastodonCache.status(for: statusID) else { fatalError("Missing cached status \(statusID)") }
        
        let account: Account
        if let reblog = status.reblog {
            account = reblog.account
        } else {
            account = status.account
        }
        self.accountID = account.id
        
        updateUIForPreferences()
        
        usernameLabel.text = "@\(account.acct)"
        avatarImageView.image = nil
        avatarURL = account.avatar
        ImageCache.avatars.get(account.avatar) { (data) in
            guard let data = data else { return }
            DispatchQueue.main.async {
                self.avatarImageView.image = UIImage(data: data)
                self.avatarURL = nil
            }
        }
        updateTimestamp()
        
        let attachments = status.attachments.filter({ $0.kind == .image })
        if attachments.count > 0 {
            attachmentsView.isHidden = false
            let width = attachmentsView.bounds.width
            let height: CGFloat = 200
            switch attachments.count {
            case 1:
                addAttachmentView(frame: CGRect(x: 0, y: 0, width: width, height: height), attachment: attachments[0])
            case 2:
                addAttachmentView(frame: CGRect(x: 0, y: 0, width: width / 2 - 4, height: height), attachment: attachments[0])
                addAttachmentView(frame: CGRect(x: width / 2 + 4, y: 0, width: width / 2 - 4, height: height), attachment: attachments[1])
            case 3:
                addAttachmentView(frame: CGRect(x: 0, y: 0, width: width / 2 - 4, height: height), attachment: attachments[0])
                addAttachmentView(frame: CGRect(x: width / 2 + 4, y: 0, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[1])
                addAttachmentView(frame: CGRect(x: width / 2 + 4, y: height / 2 + 4, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[2])
            case 4:
                addAttachmentView(frame: CGRect(x: 0, y: 0, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[0])
                addAttachmentView(frame: CGRect(x: 0, y: height / 2 + 4, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[1])
                addAttachmentView(frame: CGRect(x: width / 2 + 4, y: 0, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[2])
                addAttachmentView(frame: CGRect(x: width / 2 + 4, y: height / 2 + 4, width: width / 2 - 4, height: height / 2 - 4), attachment: attachments[3])
            default:
                fatalError("Too many attachments")
            }
        } else {
            attachmentsView.isHidden = true
        }
        
        let realStatus = status.reblog ?? status
        favorited = realStatus.favourited ?? false
        reblogged = realStatus.reblogged ?? false
        
        contentLabel.statusID = statusID
    }
    
    func updateTimestamp() {
        guard let status = MastodonCache.status(for: statusID) else { fatalError("Missing cached status \(statusID!)") }
        
        timestampLabel.text = status.createdAt.timeAgoString()
        let delay: DispatchTimeInterval?
        switch status.createdAt.timeAgo().1 {
        case .second:
            delay = .seconds(10)
        case .minute:
            delay = .seconds(60)
        default:
            delay = nil
        }
        if let delay = delay {
            updateTimestampWorkItem = DispatchWorkItem {
                self.updateTimestamp()
            }
            DispatchQueue.main.asyncAfter(deadline: .now() + delay, execute: updateTimestampWorkItem!)
        } else {
            updateTimestampWorkItem = nil
        }
    }
    
    func addAttachmentView(frame: CGRect, attachment: Attachment) {
        let attachmentView = AttachmentView(frame: frame, attachment: attachment)
        attachmentView.delegate = self
        attachmentsView.addSubview(attachmentView)
    }
    
    override func prepareForReuse() {
        if let url = avatarURL {
            ImageCache.avatars.cancel(url)
        }
        updateTimestampWorkItem?.cancel()
        updateTimestampWorkItem = nil
        attachmentsView.subviews.forEach { $0.removeFromSuperview() }
    }
    
    @objc func accountPressed() {
        delegate?.selected(account: accountID)
    }

    @IBAction func replyPressed(_ sender: Any) {
        delegate?.reply(to: statusID)
    }
    
    @IBAction func favoritePressed(_ sender: Any) {
        guard let status = MastodonCache.status(for: statusID) else { fatalError("Missing cached status \(statusID!)") }
        
        favorited = !favorited
        
        let realStatus: Status = status.reblog ?? status
        
        let request = (favorited ? Status.favourite : Status.unfavourite)(realStatus)
        MastodonController.client.run(request) { response in
            self.favorited = realStatus.favourited ?? false
            DispatchQueue.main.async {
                if case .success = response {
                    UIImpactFeedbackGenerator(style: .light).impactOccurred()
                } else {
                    print("Couldn't favorite status \(realStatus.id)")
                    // todo: display error message
                    UINotificationFeedbackGenerator().notificationOccurred(.error)
                    return
                }
            }
        }
    }
    
    @IBAction func reblogPressed(_ sender: Any) {
        guard let status = MastodonCache.status(for: statusID) else { fatalError("Missing cached status \(statusID!)") }
        
        reblogged = !reblogged
        
        let realStatus: Status = status.reblog ?? status
        
        let request = (reblogged ? Status.reblog : Status.unreblog)(realStatus)
        MastodonController.client.run(request) { response in
            self.reblogged = realStatus.reblogged ?? false
            DispatchQueue.main.async {
                if case .success = response {
                    UIImpactFeedbackGenerator(style: .light).impactOccurred()
                } else {
                    print("Couldn't reblog status \(realStatus.id)")
                    // todo: display error message
                    UINotificationFeedbackGenerator().notificationOccurred(.error)
                }
            }
        }
    }
    
    @IBAction func morePressed(_ sender: Any) {
        delegate?.showMoreOptions(forStatus: statusID)
    }
    
}

extension ConversationMainStatusTableViewCell: AttachmentViewDelegate {
    func showLargeAttachment(for attachmentView: AttachmentView) {
        if let gifData = attachmentView.gifData {
            delegate?.showLargeImage(gifData: gifData, description: attachmentView.attachment.description, animatingFrom: attachmentView)
        } else {
            delegate?.showLargeImage(attachmentView.image!, description: attachmentView.attachment.description, animatingFrom: attachmentView)
        }
    }
}

extension ConversationMainStatusTableViewCell: PreviewViewControllerProvider {
    func getPreviewViewController(forLocation location: CGPoint, sourceViewController: UIViewController) -> UIViewController? {
        if avatarImageView.frame.contains(location) {
            return ProfileTableViewController(accountID: accountID)
        } else if attachmentsView.frame.contains(location) {
            let attachmentsViewLocation = attachmentsView.convert(location, from: self)
            if let attachmentView = attachmentsView.subviews.first(where: { $0.frame.contains(attachmentsViewLocation) }) as? AttachmentView {
                let image = attachmentView.image!
                let description = attachmentView.description
                return delegate?.largeImage(image, description: description, sourceView: attachmentView)
            }
        } else if contentLabel.frame.contains(location),
            let vc = contentLabel.getViewController(forLinkAt: contentLabel.convert(location, from: self)) {
            return vc
        }
        return nil
    }
}
